@@ -5,7 +5,7 @@ import * as vscode from 'vscode';
 import { SymbolKind, commands } from 'vscode';
 import { Disposable } from 'vscode-jsonrpc';
 import { ExecuteCommandRequest } from 'vscode-languageclient';
-import { ExtensionState } from './ExtensionState';
+import { ALSSourceDirDescription, ExtensionState } from './ExtensionState';
 import { AdaConfig, getOrAskForProgram, initializeConfig } from './debugConfigProvider';
 import { adaExtState, logger, mainOutputChannel } from './extension';
 import { findAdaMain, getProjectFileRelPath, getSymbols } from './helpers';
@@ -18,12 +18,17 @@ import {
     TASK_PROVE_SUPB_PLAIN_NAME,
     TASK_TYPE_SPARK,
     findBuildAndRunTask,
-    getBuildAndRunTasks,
+    getTasksWithPrefix,
     getConventionalTaskLabel,
     isFromWorkspace,
     workspaceTasksFirst,
+    getBuildAndRunTaskName,
+    getRunGNATemulatorTaskName,
+    getBuildTaskName,
+    runTaskAndGetResult,
 } from './taskProviders';
 import { createHelloWorldProject, walkthroughStartDebugging } from './walkthrough';
+import { loadGnatCoverageReport } from './gnattest';
 
 /**
  * Identifier for a hidden command used for building and running a project main.
@@ -43,6 +48,26 @@ export const CMD_BUILD_AND_RUN_MAIN = 'ada.buildAndRunMain';
  */
 export const CMD_BUILD_AND_DEBUG_MAIN = 'ada.buildAndDebugMain';
 
+/**
+ * Identifier for a hidden command used for building and running a project main,
+ * using GNATemulator.
+ * The command accepts a parameter which is the URI of the main source file.
+ * It is triggered by CodeLenses provided by the extension.
+ *
+ * @see {@link buildAndRunMainWithGNATemulator}
+ */
+export const CMD_BUILD_AND_RUN_GNATEMULATOR = 'ada.buildAndRunGNATemulator';
+
+/**
+ * Identifier for a hidden command used for building and debugging a project main,
+ * using GNATemulator.
+ * The command accepts a parameter which is the URI of the main source file.
+ * It is triggered by CodeLenses provided by the extension.
+ *
+ * @see {@link buildAndDebugSpecifiedMain}
+ */
+export const CMD_BUILD_AND_DEBUG_GNATEMULATOR = 'ada.buildAndDebugGNATemulator';
+
 /**
  * Identifier for a hidden command that returns an array of strings constituting
  * the -P and -X project and scenario arguments.
@@ -69,6 +94,21 @@ export function registerCommands(context: vscode.ExtensionContext, clients: Exte
         vscode.commands.registerCommand('ada.walkthroughStartDebugging', walkthroughStartDebugging),
     );
     context.subscriptions.push(vscode.commands.registerCommand('ada.otherFile', otherFileHandler));
+    context.subscriptions.push(
+        vscode.commands.registerCommand('ada.createNewAdaMainUnit', () =>
+            createNewFile('ada', 'Main Procedure'),
+        ),
+    );
+    context.subscriptions.push(
+        vscode.commands.registerCommand('ada.createNewAdaPackage', () =>
+            createNewFile('ada', 'Package Declaration or Body'),
+        ),
+    );
+    context.subscriptions.push(
+        vscode.commands.registerCommand('ada.createNewGPRProjectFile', () =>
+            createNewFile('gpr', 'Project Declaration'),
+        ),
+    );
     context.subscriptions.push(
         vscode.commands.registerCommand('ada.subprogramBox', addSubprogramBoxCommand),
     );
@@ -127,7 +167,18 @@ export function registerCommands(context: vscode.ExtensionContext, clients: Exte
     context.subscriptions.push(
         vscode.commands.registerCommand(CMD_BUILD_AND_DEBUG_MAIN, buildAndDebugSpecifiedMain),
     );
-
+    context.subscriptions.push(
+        vscode.commands.registerCommand(
+            CMD_BUILD_AND_RUN_GNATEMULATOR,
+            buildAndRunMainWithGNATemulator,
+        ),
+    );
+    context.subscriptions.push(
+        vscode.commands.registerCommand(
+            CMD_BUILD_AND_DEBUG_GNATEMULATOR,
+            buildAndDebugSpecifiedMainWithGNATemulator,
+        ),
+    );
     context.subscriptions.push(
         vscode.commands.registerCommand(CMD_GPR_PROJECT_ARGS, gprProjectArgs),
     );
@@ -144,6 +195,10 @@ export function registerCommands(context: vscode.ExtensionContext, clients: Exte
         vscode.commands.registerCommand(CMD_SPARK_PROVE_SUBP, sparkProveSubprogram),
     );
 
+    context.subscriptions.push(
+        commands.registerCommand('ada.loadGnatCovXMLReport', loadGnatCovXMLReport),
+    );
+
     registerTaskWrappers(context);
 }
 
@@ -241,7 +296,7 @@ let lastUsedTaskInfo: { source: string; name: string } | undefined;
  * @returns the TaskExecution corresponding to the task.
  */
 async function buildAndRunMainLast() {
-    const buildAndRunTasks = await getBuildAndRunTasks();
+    const buildAndRunTasks = await getTasksWithPrefix(getBuildAndRunTaskName());
     if (lastUsedTaskInfo) {
         const matchingTasks = buildAndRunTasks.filter(matchesLastUsedTask);
         assert(matchingTasks.length <= 1);
@@ -305,7 +360,7 @@ async function buildAndRunMainAsk() {
             ],
         };
     }
-    const adaTasksMain = await getBuildAndRunTasks();
+    const adaTasksMain = await getTasksWithPrefix(getBuildAndRunTaskName());
 
     if (adaTasksMain.length > 0) {
         const tasksFromWorkspace = adaTasksMain.filter(isFromWorkspace);
@@ -412,6 +467,24 @@ async function buildAndRunMainAsk() {
     }
 }
 
+/**
+ * Handler for commands that create new files.
+ * This function creates a new editor for the given language, focus it,
+ * and insert the specified snippet.
+ * Used to proivide Ada/GPR file templates.
+ *
+ * @param langId - the new file's language ID (e.g: 'ada' or 'gpr')
+ * @param snippetName - the name of the snippet to insert in the newly created editor.
+ */
+async function createNewFile(langId: string, snippetName: string) {
+    const doc = await vscode.workspace.openTextDocument({ language: 'ada' });
+    await vscode.window.showTextDocument(doc);
+    await vscode.commands.executeCommand('editor.action.insertSnippet', {
+        langId: langId,
+        name: snippetName,
+    });
+}
+
 //  Take active editor URI and call execute 'als-other-file' command in LSP
 const otherFileHandler = () => {
     const activeEditor = vscode.window.activeTextEditor;
@@ -441,13 +514,7 @@ const otherFileHandler = () => {
  * when missing directories
  */
 export async function checkSrcDirectories(atStartup = false, displayYesNoPopup = true) {
-    type ALSSourceDirDescription = {
-        name: string;
-        uri: string;
-    };
-
     const foldersInSettings = vscode.workspace.getConfiguration().get('folders');
-    const alsClient = adaExtState.adaClient;
     const doNotShowAgainKey = 'ada.addMissingDirsToWorkspace.doNotShowAgain';
     const doNotShowAgain = adaExtState.context.workspaceState.get(doNotShowAgainKey);
 
@@ -456,13 +523,7 @@ export async function checkSrcDirectories(atStartup = false, displayYesNoPopup =
     //  triggered at startup while the user previously clicked on the
     //  'Don't show again' button for this workspace
     if (foldersInSettings === undefined && !(atStartup && doNotShowAgain)) {
-        const sourceDirs: ALSSourceDirDescription[] = (await alsClient.sendRequest(
-            ExecuteCommandRequest.type,
-            {
-                command: 'als-source-dirs',
-            },
-        )) as ALSSourceDirDescription[];
-
+        const sourceDirs: ALSSourceDirDescription[] = await adaExtState.getSourceDirs();
         const isSubdirectory = (dir: string, parent: string) => {
             //  Use lower-case on Windows since drives can be specified in VS Code
             //  either with lower or upper case characters.
@@ -555,17 +616,23 @@ export async function checkSrcDirectories(atStartup = false, displayYesNoPopup =
  * displayed.
  *
  * @param main - a URI of a document
+ * @param useGNATemulator - whether the main should be ran through
+ * GNATemulator.
  */
-async function buildAndRunSpecifiedMain(main: vscode.Uri): Promise<void> {
+async function buildAndRunSpecifiedMain(
+    main: vscode.Uri,
+    useGNATemulator: boolean = false,
+): Promise<void> {
     const adaMain = await findAdaMain(main.fsPath);
     if (adaMain) {
-        const task = await findBuildAndRunTask(adaMain);
+        const task = await findBuildAndRunTask(adaMain, useGNATemulator);
         if (task) {
             lastUsedTaskInfo = { source: task.source, name: task.name };
             await vscode.tasks.executeTask(task);
         } else {
+            const taskLabel = useGNATemulator ? 'Build and Run GNATemulator' : 'Build and Run';
             void vscode.window.showErrorMessage(
-                `Could not find the 'Build and Run' task for the project main ` +
+                `Could not find the '${taskLabel}' task for the project main ` +
                     `${adaMain.mainRelPath()}`,
                 { modal: true },
             );
@@ -579,20 +646,39 @@ async function buildAndRunSpecifiedMain(main: vscode.Uri): Promise<void> {
     }
 }
 
-/**
- * This is a command handler that builds the main given as parameter and starts
- * a debug session on that main.  If the given URI does not match one of the
- * project Mains an error is displayed.
+/*
+ * This is a command handler that builds and runs the main given as parameter,
+ * using GNATemulator to run the executable.
+ * If the given URI does not match one of the project Mains an error is
+ * displayed.
  *
  * @param main - a URI of a document
  */
-async function buildAndDebugSpecifiedMain(main: vscode.Uri): Promise<void> {
+async function buildAndRunMainWithGNATemulator(main: vscode.Uri): Promise<void> {
+    return buildAndRunSpecifiedMain(main, true);
+}
+
+/**
+ * This is a command handler that builds the main given as parameter and starts
+ * a debug session on that main, running GNATemulator in debug mode before when
+ * asked.
+ * If the given URI does not match one of the project Mains an error is displayed.
+ *
+ * @param main - a URI of a document
+ * @param useGNATemulator - whether the main should be ran through GNATemulator
+ */
+async function buildAndDebugSpecifiedMain(
+    main: vscode.Uri,
+    useGNATemulator: boolean = false,
+): Promise<void> {
     function isMatchingConfig(cfg: vscode.DebugConfiguration, configToMatch: AdaConfig): boolean {
         return cfg.type == configToMatch.type && cfg.name == configToMatch.name;
     }
 
     const wsFolder = vscode.workspace.getWorkspaceFolder(main);
     const adaMain = await findAdaMain(main.fsPath);
+    const target = await adaExtState.getTargetPrefix();
+    const debugServerAddress = await adaExtState.getDebugServerAddress(useGNATemulator);
     if (adaMain) {
         /**
          * The vscode API doesn't provide a way to list both automatically
@@ -602,7 +688,7 @@ async function buildAndDebugSpecifiedMain(main: vscode.Uri): Promise<void> {
          * the given main URI.
          */
         // Create a launch config for this main to help with matching
-        const configToMatch = initializeConfig(adaMain);
+        const configToMatch = initializeConfig(adaMain, target, debugServerAddress);
         logger.debug('Debug config to match:\n' + JSON.stringify(configToMatch, null, 2));
 
         let matchingConfig = undefined;
@@ -628,6 +714,38 @@ async function buildAndDebugSpecifiedMain(main: vscode.Uri): Promise<void> {
 
         if (matchingConfig) {
             logger.debug('Found matching config: ' + JSON.stringify(matchingConfig, null, 2));
+
+            // Trying to debug via GNATemulator: run the main via GNATemulator in debug mode
+            // before starting the debug session
+            if (useGNATemulator) {
+                const buildTaskName = getBuildTaskName(adaMain);
+                const buildTasks = await getTasksWithPrefix(buildTaskName);
+                if (buildTasks.length === 1) {
+                    const execStatus: number | undefined = await runTaskAndGetResult(buildTasks[0]);
+                    if (execStatus != 0) {
+                        const errorMsg = `Failed to build executable before launching GNATemulator`;
+                        logger.error(errorMsg);
+                        return;
+                    }
+                    logger.debug('Running GNATemulator before starting the debug configuration...');
+                    const runGNATemulatorTaskName = getRunGNATemulatorTaskName(adaMain, true);
+                    const runGNATemulatorForDebugTasks =
+                        await getTasksWithPrefix(runGNATemulatorTaskName);
+
+                    if (runGNATemulatorForDebugTasks.length < 1) {
+                        const errorMsg = `Could not find '${runGNATemulatorTaskName}' task`;
+                        logger.error(errorMsg);
+                        void vscode.window.showErrorMessage(errorMsg, { modal: true });
+                        return;
+                    }
+                    await vscode.tasks.executeTask(runGNATemulatorForDebugTasks[0]);
+                } else {
+                    const errorMsg = `Could not find '${buildTaskName}' task`;
+                    void vscode.window.showErrorMessage(errorMsg, { modal: true });
+                    logger.error(errorMsg);
+                }
+            }
+
             const success = await vscode.debug.startDebugging(wsFolder, matchingConfig);
             if (!success) {
                 void vscode.window.showErrorMessage(
@@ -651,21 +769,41 @@ async function buildAndDebugSpecifiedMain(main: vscode.Uri): Promise<void> {
     }
 }
 
+/**
+ * This is a command handler that builds the main given as parameter and starts
+ * a debug session on that main, with GNATemulator.
+ * If the given URI does not match one of the project Mains an error is displayed.
+ *
+ * @param main - a URI of a document
+ */
+async function buildAndDebugSpecifiedMainWithGNATemulator(main: vscode.Uri): Promise<void> {
+    await buildAndDebugSpecifiedMain(main, true);
+}
+
 /**
  * @returns an array of -P and -X project and scenario command lines arguments
  * for use with GPR-based tools.
  */
 export async function gprProjectArgs(): Promise<string[]> {
-    const vars: string[][] = Object.entries(
-        vscode.workspace.getConfiguration('ada').get('scenarioVariables') ?? [],
-    );
-    return ['-P', await getProjectFromConfigOrALS()].concat(
-        vars.map(([key, value]) => `-X${key}=${value}`),
-    );
+    const scenarioArgs = gprScenarioArgs();
+
+    return ['-P', await getProjectFromConfigOrALS()].concat(scenarioArgs);
 }
 
 export const PROJECT_FROM_CONFIG = '${config:ada.projectFile}';
 
+/**
+ * @returns an array of -X scenario command lines arguments for use with
+ * GPR-based tools.
+ */
+export function gprScenarioArgs() {
+    const vars: string[][] = Object.entries(
+        vscode.workspace.getConfiguration('ada').get('scenarioVariables') ?? [],
+    );
+    const scenarioArgs = vars.map(([key, value]) => `-X${key}=${value}`);
+    return scenarioArgs;
+}
+
 /**
  * @returns `"$\{config:ada.projectFile\}"` if that setting has a value, or else
  * queries the ALS for the current project and returns the full path.
@@ -877,3 +1015,24 @@ async function sparkProveSubprogram(
      */
     return await vscode.tasks.executeTask(resolvedTask);
 }
+async function loadGnatCovXMLReport() {
+    const selection = await vscode.window.showOpenDialog({
+        canSelectFiles: true,
+        canSelectFolders: false,
+        canSelectMany: false,
+        filters: {
+            'index.xml': ['xml'],
+        },
+        title: "Select a 'index.xml' GNATcoverage report to load",
+    });
+
+    if (selection && selection.length > 0 && selection[0]) {
+        const path = selection[0].fsPath;
+        if (!path.endsWith('index.xml')) {
+            throw Error(
+                `The selected file must be 'index.xml'. Instead, the selected file was: ${path}`,
+            );
+        }
+        await loadGnatCoverageReport(path);
+    }
+}
