@@ -16,10 +16,12 @@
 ----------------------------------------------------------------------------*/
 
 import assert from 'assert';
+import { existsSync } from 'fs';
 import path, { basename } from 'path';
 import * as vscode from 'vscode';
 import { CMD_GPR_PROJECT_ARGS } from './commands';
 import { adaExtState, logger } from './extension';
+import { getGnatTestDriverProjectPath } from './gnattest';
 import { AdaMain, getAdaMains, showErrorMessageWithOpenLogButton } from './helpers';
 
 export const TASK_TYPE_ADA = 'ada';
@@ -57,6 +59,8 @@ interface PredefinedTask {
     taskDef: SimpleTaskDef;
     problemMatchers?: string | string[];
     taskGroup?: vscode.TaskGroup;
+    revealKind?: vscode.TaskRevealKind;
+    isBackground?: boolean;
 }
 
 /**
@@ -90,6 +94,19 @@ export const TASK_PROVE_SUPB_PLAIN_NAME = 'Prove subprogram';
 export const TASK_PROVE_REGION_PLAIN_NAME = 'Prove selected region';
 export const TASK_PROVE_LINE_PLAIN_NAME = 'Prove line';
 export const TASK_PROVE_FILE_PLAIN_NAME = 'Prove file';
+export const TASK_BUILD_TEST_DRIVER = 'Build GNATtest test harness project';
+
+export const TASK_GNATCOV_SETUP: PredefinedTask = {
+    label: 'GNATcoverage - Setup runtime library',
+    taskDef: {
+        type: TASK_TYPE_ADA,
+        command: 'gnatcov',
+        args: ['setup'],
+    },
+    problemMatchers: DEFAULT_PROBLEM_MATCHER,
+};
+
+const gnatCovTasks: PredefinedTask[] = [TASK_GNATCOV_SETUP];
 
 /**
  * Predefined tasks offered by the extension. Both 'ada' and 'spark' tasks are
@@ -230,7 +247,7 @@ const predefinedTasks: PredefinedTask[] = [
         problemMatchers: '',
     },
     {
-        label: 'Create/update test skeletons for the project',
+        label: 'Create or update GNATtest test framework',
         taskDef: {
             type: TASK_TYPE_ADA,
             command: 'gnattest',
@@ -366,7 +383,7 @@ const predefinedTasks: PredefinedTask[] = [
         },
         problemMatchers: DEFAULT_PROBLEM_MATCHER,
     },
-];
+].concat(gnatCovTasks);
 
 /**
  * A provider of tasks based on the {@link SimpleTaskDef} task definition.
@@ -387,6 +404,8 @@ export class SimpleTaskProvider implements vscode.TaskProvider {
         }
 
         const result: vscode.Task[] = [];
+        const targetPrefix = await adaExtState.getTargetPrefix();
+        const isNativeProject = await adaExtState.isNativeProject();
 
         /**
          * Start with the list of predefined tasks.
@@ -440,6 +459,7 @@ export class SimpleTaskProvider implements vscode.TaskProvider {
                             command: execRelPath,
                             args: [],
                         },
+                        revealKind: vscode.TaskRevealKind.Always,
                     };
 
                     const buildAndRunTask: PredefinedTask = {
@@ -451,9 +471,93 @@ export class SimpleTaskProvider implements vscode.TaskProvider {
                         problemMatchers: '',
                     };
 
-                    return [buildTask, runTask, buildAndRunTask];
+                    const tasks = [buildTask, runTask, buildAndRunTask];
+
+                    if (!isNativeProject) {
+                        const gnatemulatorExe = targetPrefix
+                            ? targetPrefix + '-gnatemu'
+                            : 'gnatemu';
+                        const runGNATemulatorTask: PredefinedTask = {
+                            label: getRunGNATemulatorTaskPlainName(main),
+                            taskDef: {
+                                type: this.taskType,
+                                command: gnatemulatorExe,
+                                args: [`\${command:${CMD_GPR_PROJECT_ARGS}}`, execRelPath],
+                            },
+                            revealKind: vscode.TaskRevealKind.Always,
+                        };
+
+                        const runGNATemulatorTaskForDebug: PredefinedTask = {
+                            ...runGNATemulatorTask,
+                            label: getRunGNATemulatorTaskPlainName(main, true),
+                            taskDef: {
+                                ...runGNATemulatorTask.taskDef,
+                                args: [`\${command:${CMD_GPR_PROJECT_ARGS}}`, `-g`, execRelPath],
+                            },
+                            isBackground: true,
+                            // This task is run before some debug tasks.
+                            // Problem is, it's a watch script, and since it never exits, VSCode
+                            // complains. All this is needed so VSCode just lets it run.
+                            problemMatchers: '',
+                        };
+
+                        const buildAndRunGNATemulatorTask: PredefinedTask = {
+                            label: getBuildAndRunGNATemulatorTaskPlainName(main),
+                            taskDef: {
+                                type: this.taskType,
+                                compound: [buildTask.label, runGNATemulatorTask.label],
+                            },
+                            problemMatchers: '',
+                        };
+
+                        const buildAndRunGNATemulatorTaskForDebug: PredefinedTask = {
+                            label: getBuildAndRunGNATemulatorTaskPlainName(main, true),
+                            taskDef: {
+                                type: this.taskType,
+                                compound: [buildTask.label, runGNATemulatorTaskForDebug.label],
+                            },
+                            problemMatchers: '',
+                        };
+
+                        tasks.push(
+                            runGNATemulatorTask,
+                            runGNATemulatorTaskForDebug,
+                            buildAndRunGNATemulatorTask,
+                            buildAndRunGNATemulatorTaskForDebug,
+                        );
+                    }
+
+                    return tasks;
                 }),
             );
+
+            /**
+             * If a test harness project exists, provide a task to build it.
+             */
+            const harnessPrj = await getGnatTestDriverProjectPath().catch(
+                /**
+                 * In case of errors fallback silently to keep providing tasks
+                 * even in case of errors.
+                 */
+                (reason) => {
+                    logger.error(
+                        'Error while querying for the GNATtest test driver project:\n' + reason,
+                    );
+                    return undefined;
+                },
+            );
+            if (harnessPrj && existsSync(harnessPrj)) {
+                taskDeclsToOffer.push({
+                    label: TASK_BUILD_TEST_DRIVER,
+                    taskDef: {
+                        type: TASK_TYPE_ADA,
+                        command: 'gprbuild',
+                        args: ['-P', harnessPrj, "'-cargs:ada'", '-gnatef'],
+                    },
+                    problemMatchers: DEFAULT_PROBLEM_MATCHER,
+                    taskGroup: vscode.TaskGroup.Build,
+                });
+            }
         }
 
         /**
@@ -499,6 +603,17 @@ export class SimpleTaskProvider implements vscode.TaskProvider {
                     resolvedTask.group = tDecl.taskGroup;
                 }
 
+                /**
+                 * If the task has a specified revealKind, use it.
+                 * Otherwise, by default, do not reveal the Terminal panel when
+                 * running the task: we want to highlight the issues displayed
+                 * in the Problems view instead.
+                 */
+                resolvedTask.presentationOptions = {
+                    reveal: tDecl.revealKind ?? vscode.TaskRevealKind.Never,
+                };
+                resolvedTask.isBackground = tDecl.isBackground ?? false;
+
                 result.push(resolvedTask);
             } else {
                 logger.error(`Failed to resolve task: ${JSON.stringify(task, undefined, 2)}`);
@@ -750,6 +865,21 @@ function getRunTaskPlainName(main?: AdaMain) {
     return `Run main - ${main?.mainRelPath() ?? ''}`;
 }
 
+/**
+ * The full name of the 'Run GNATemulator' task of a main, including the task type.
+ */
+export function getRunGNATemulatorTaskName(main?: AdaMain, forDebug = false) {
+    return `${TASK_TYPE_ADA}: ${getRunGNATemulatorTaskPlainName(main, forDebug)}`;
+}
+
+/**
+ * The name of the 'Run GNATemulator' task of a main, without the task type.
+ */
+function getRunGNATemulatorTaskPlainName(main?: AdaMain, forDebug = false) {
+    const forDebugLabelSuffix = forDebug ? ' (debug)' : '';
+    return `Run main with GNATemulator${forDebugLabelSuffix} - ${main?.mainRelPath() ?? ''}`;
+}
+
 /**
  * The full name of the build task of a main, including the task type.
  */
@@ -761,10 +891,22 @@ export function getBuildAndRunTaskPlainName(main?: AdaMain) {
     return `Build and run main - ${main?.mainRelPath() ?? ''}`;
 }
 
+export function getBuildAndRunGNATemulatorTaskPlainName(main?: AdaMain, forDebug = false) {
+    const forDebugLabelSuffix = forDebug ? ' (debug)' : '';
+    return (
+        `Build and run main with GNATemulator` +
+        `${forDebugLabelSuffix} - ${main?.mainRelPath() ?? ''}`
+    );
+}
+
 export function getBuildAndRunTaskName(main?: AdaMain) {
     return `${TASK_TYPE_ADA}: ${getBuildAndRunTaskPlainName(main)}`;
 }
 
+export function getBuildAndRunGNATemulatorTaskName(main?: AdaMain, forDebug = false) {
+    return `${TASK_TYPE_ADA}: ${getBuildAndRunGNATemulatorTaskPlainName(main, forDebug)}`;
+}
+
 export function createSparkTaskProvider(): SimpleTaskProvider {
     return new SimpleTaskProvider(
         TASK_TYPE_SPARK,
@@ -870,7 +1012,12 @@ abstract class SequentialExecution extends vscode.CustomExecution {
                                 try {
                                     if (reason instanceof Error) {
                                         void vscode.window.showErrorMessage(reason.message);
-                                        writeEmitter.fire(reason.message + '\r\n');
+                                        /**
+                                         * Emit lines with \\r\\n because we're writing to a console
+                                         */
+                                        reason.message
+                                            .split(/\r?\n|\r|\n/g)
+                                            .forEach((line) => writeEmitter.fire(line + '\r\n'));
                                     }
                                 } finally {
                                     closeEmitter.fire(2);
@@ -968,9 +1115,9 @@ class SequentialExecutionByName extends SequentialExecution {
  * @param tasks - list of tasks to run in sequence.
  * @returns Status of the last executed task.
  */
-function runTaskSequence(
+export function runTaskSequence(
     tasks: vscode.Task[],
-    writeEmitter: vscode.EventEmitter<string>,
+    writeEmitter?: vscode.EventEmitter<string>,
 ): Promise<number> {
     let p = new Promise<number>((resolve) => resolve(0));
     for (const t of tasks) {
@@ -984,9 +1131,9 @@ function runTaskSequence(
                         }
                     });
 
-                    writeEmitter.fire(`Executing task: ${getConventionalTaskLabel(t)}\r\n`);
+                    writeEmitter?.fire(`Executing task: ${getConventionalTaskLabel(t)}\r\n`);
                     vscode.tasks.executeTask(t).then(undefined, (reason) => {
-                        writeEmitter.fire(`Could not execute task: ${reason}\r\n`);
+                        writeEmitter?.fire(`Could not execute task: ${reason}\r\n`);
                         // eslint-disable-next-line @typescript-eslint/prefer-promise-reject-errors
                         reject(reason);
                     });
@@ -1014,27 +1161,33 @@ export const workspaceTasksFirst = (a: vscode.Task, b: vscode.Task): number => {
 
 /**
  *
- * @returns Array of tasks of type `ada` and label starting with 'Build and run
- * main -'. This includes tasks automatically provided by the extension as well
+ * @returns Array of tasks of type `ada` and label starting with the given prefix.
+ * This includes tasks automatically provided by the extension as well
  * as user-defined tasks in tasks.json.
  */
-export async function getBuildAndRunTasks(): Promise<vscode.Task[]> {
+export async function getTasksWithPrefix(prefix: string): Promise<vscode.Task[]> {
     return await vscode.tasks.fetchTasks({ type: TASK_TYPE_ADA }).then((tasks) =>
         tasks
-            // Filter to tasks starting with the conventional name of "Build and run main" tasks
-            .filter((t) => getConventionalTaskLabel(t).startsWith(getBuildAndRunTaskName()))
+            // Filter to tasks starting with a conventional name that has the given
+            // prefix.
+            .filter((t) => getConventionalTaskLabel(t).startsWith(prefix))
 
             // Return workspace-defined tasks first
             .sort(workspaceTasksFirst),
     );
 }
 
-export async function findBuildAndRunTask(adaMain: AdaMain): Promise<vscode.Task | undefined> {
-    return (await getBuildAndRunTasks()).find(
-        // Tasks defined in tasks.json will have a leading 'ada: ' while the
-        // ones auto-generated by the extension don't. We want to match both.
-        (t) => getConventionalTaskLabel(t) == getBuildAndRunTaskName(adaMain),
-    );
+export async function findBuildAndRunTask(
+    adaMain: AdaMain,
+    useGNATemulator: boolean = false,
+): Promise<vscode.Task | undefined> {
+    const label = useGNATemulator
+        ? getBuildAndRunGNATemulatorTaskName(adaMain)
+        : getBuildAndRunTaskName(adaMain);
+
+    const tasks = await getTasksWithPrefix(label);
+
+    return tasks.length == 1 ? tasks[0] : undefined;
 }
 
 /**
@@ -1148,3 +1301,63 @@ function isAlire(command: string | vscode.ShellQuotedString): boolean {
     const commandBasename = basename(value);
     return commandBasename.match(/^alr(\.exe)?$/) != null;
 }
+
+/**
+ * Execute the given task, wait until it finishes and return the underlying
+ * process exit code.
+ *
+ * @param task - a {@link vscode.Task}
+ * @returns a Promise that resolves to the underlying process exit code when the
+ * task finishes execution.
+ */
+export async function runTaskAndGetResult(task: vscode.Task): Promise<number | undefined> {
+    return await new Promise<number | undefined>((resolve, reject) => {
+        let started = false;
+
+        const startDisposable = vscode.tasks.onDidStartTask((e) => {
+            if (e.execution.task == task) {
+                /**
+                 * Task was started, let's listen to the end.
+                 */
+                started = true;
+                startDisposable.dispose();
+            }
+        });
+
+        const endDisposable = vscode.tasks.onDidEndTaskProcess((e) => {
+            if (e.execution.task == task) {
+                endDisposable.dispose();
+                resolve(e.exitCode);
+            }
+        });
+
+        setTimeout(() => {
+            /**
+             * If the task has not started within the timeout below, it means an
+             * error occured during startup. Reject the promise.
+             */
+            if (!started) {
+                const msg = `The task '${getConventionalTaskLabel(
+                    task,
+                )}' was not started, likely due to an error.\n`;
+                reject(Error(msg));
+            }
+        }, 3000);
+
+        void vscode.tasks.executeTask(task);
+    }).catch(async (reason) => {
+        if (reason instanceof Error) {
+            let msg = 'The current list of tasks is:\n';
+            msg += await vscode.tasks.fetchTasks({ type: task.definition.type }).then(
+                (list) => list.map(getConventionalTaskLabel).join('\n'),
+
+                (reason) => `fetchTasks promise was rejected: ${reason}`,
+            );
+
+            reason.message += '\n' + msg;
+        }
+
+        // eslint-disable-next-line @typescript-eslint/prefer-promise-reject-errors
+        return Promise.reject(reason);
+    });
+}
