@@ -6,6 +6,7 @@ import inspect
 import json
 import logging
 import os
+import psutil
 import shlex
 import sys
 import urllib
@@ -222,7 +223,7 @@ class ALSSettings(
     defaultCharset: str | None
     displayMethodAncestryOnNavigation: bool | None
     documentationStyle: Literal["gnat", "leading"] | None
-    enableDiagnostics: bool | None
+    adaFileDiagnostics: bool | None
     enableIndexing: bool | None
     foldComments: bool | None
     followSymlinks: bool | None
@@ -451,6 +452,16 @@ class ALSLanguageClient(LanguageClient):
             )
         )
 
+    async def references(self, uri: str, line_one_based: int, char_one_based: int):
+        pos = Pos(line_one_based, char_one_based)
+        return await self.text_document_references_async(
+            lsprotocol.types.ReferenceParams(
+                context=lsprotocol.types.ReferenceContext(include_declaration=True),
+                text_document=TextDocumentIdentifier(uri),
+                position=pos,
+            )
+        )
+
     def assertEqual(self, actual: Any, expected: Any) -> None:
         """Raise an AssertionError if actual != expected."""
         assertEqual(actual, expected)
@@ -784,12 +795,12 @@ def do_main():
     run_test_file(args.test_py_path)
 
 
-def Pos(line_one_based: int, char_one_based: int):
+def Pos(line_one_based: int, char_one_based: int) -> Position:
     """Shortcut for creating a Position object with ONE-BASED locations."""
     return Position(line_one_based - 1, char_one_based - 1)
 
 
-def RangeZero(line_one_based: int, char_one_based: int):
+def RangeZero(line_one_based: int, char_one_based: int) -> Range:
     """Shortcut for creating a Range that starts and ends at the same location given
     with ONE-BASED integers.
     """
@@ -835,6 +846,19 @@ def URI(src_path: Path | str) -> str:
     return src_uri
 
 
+def find_ALS_process():
+    """
+    Return the ALS process launched by the pyslsp test driver.
+    """
+    for x in psutil.pids():
+        try:
+            p = psutil.Process(x)
+            if p.ppid() == os.getpid() and p.name().startswith("ada_language_server"):
+                return p
+        except psutil.NoSuchProcess:
+            pass
+
+
 def assertEqual(actual: Any, expected: Any) -> None:
     """Raise an AssertionError if actual != expected."""
     if actual != expected:
@@ -868,6 +892,20 @@ def assertLocationsList(
     )
 
 
+def get_ALS_memory_footprint():
+    """
+    Return the memory footprint in MB of the ALS process launched
+    by the pyslsp test driver.
+    """
+    for x in psutil.pids():
+        try:
+            p = psutil.Process(x)
+            if p.ppid() == os.getpid() and p.name().startswith("ada_language_server"):
+                return p.memory_full_info().rss / (1024**2)
+        except psutil.NoSuchProcess:
+            pass
+
+
 class TestInfraError(Exception):
 
     def __init__(
